import bpy
import bpy.ops
import copy
import mathutils
import bmesh
import bmesh.ops
import os
from time import time
from math import radians, pi, ceil
from random import seed, random, randint, uniform
from mathutils import Vector, Quaternion
from bpy.props import FloatProperty, BoolProperty, IntProperty, StringProperty

create_cylinder = bpy.ops.mesh.primitive_cylinder_add
create_plane = bpy.ops.mesh.primitive_plane_add
scale = bmesh.ops.scale

bl_info = {
    "name": "Tree Creator",
    "author": "Student 14027009",
    "version": (1, 0, 0),
    "location": "View 3D > Object Mode > Tool Shelf",
    "description": "Generate a tree mesh from information supplied in the properties pane",
    "category": "Add Mesh"
}


# ------------------------------------------------------------ #
#                           UTILS                              #
# ------------------------------------------------------------ #
def load_texture(filepath):
    """Loads the texture at filepath then returns the loaded image"""
    try:
        img = bpy.data.images.load(filepath)
        return img
    except:
        raise NameError("No image was found at this location: {}".format(filepath))


def find_upper_face(bm, lowest=True):
    """
    This function will return the face which is at the
    lowest location in the mesh by default, otherwise it will
    return the highest face in the z axis

    Variables:
    :param bm: BMesh for the object
    :param lowest: A boolean whether or not it should return the lower face or the higher face

    :return a BMFace object
    """
    lowest_face = None
    lowest_face_location = None
    for face in bm.faces:
        location = face.calc_center_median()
        if lowest_face:
            # Search for lowest face in z
            if lowest:
                if lowest_face_location.z > location.z:
                    lowest_face = face
                    lowest_face_location = location
            # Search for highest face in z
            else:
                if lowest_face_location.z < location.z:
                    lowest_face = face
                    lowest_face_location = location
        # Set the first lowestFace
        else:
            lowest_face = face
            lowest_face_location = location
    return lowest_face


def scale_objects_to_match(object1, object2):
    """
    This function will take object1 and scale it up so that it
    will fit onto object2, note that the scaling will only scale
    the face that is at the top of the mesh. This means that this
    function is only really meant for objects directly ontop of
    each other.

    Note that this will only work with an object which has a single
    flat face at the top

    Variables:
    :param object1: This is the object that will be scaled to the other object
    :param object2: This is the object that provides the dimensions

    :return None
    """
    bm1 = bmesh.new()
    bm2 = bmesh.new()
    bm1.from_mesh(object1.data)
    bm2.from_mesh(object2.data)
    face1 = find_upper_face(bm1, lowest=False)
    face2 = find_upper_face(bm2)
    area1 = face1.calc_area() * object1.scale
    area2 = face2.calc_area() * object2.scale
    ratio_x = area2.x / area1.x
    ratio_y = area2.y / area1.y

    # Half the effect of the scaling 
    ratio_x = 1 - ((1 - ratio_x) / 2)
    ratio_y = 1 - ((1 - ratio_y) / 2)
    scale(bm1, vec=mathutils.Vector((ratio_x, ratio_y, 1)), verts=face1.verts)
    bm1.to_mesh(object1.data)
    return None


def join_objects_by_name(name):
    """
    Select all the objects by name and join all of the objects around the last
    selected object

    Variables:
    :param name: The base string name of the objects to join

    :return the newly created object
    """
    bpy.ops.object.select_pattern(extend=False, pattern=name + "*",
                                  case_sensitive=False)
    bpy.context.scene.objects.active = bpy.context.selected_objects[-1]
    bpy.ops.object.join()
    obj = bpy.context.scene.objects.active
    return obj


def get_cylinder_top(obj, base_position):
    """
    Calculates the point at the top of a cylinder object given the
    object and base_position

    Variables:
    :param obj: this is the cylinder that needs
    :param base_position: this is the position that is at the bottom of the cylinder

    :return The point at the top of the cylinder
    """
    center = obj.location
    offset = center - base_position
    top = center + offset
    return top


def get_bottom_bounds(obj):
    """
    This function will return the bottom bounding points for the
    given object

    Variables:
    :param obj: This is the object that will have its bottom bounds returned

    :return a list of the bottom bounding corners
    """
    bounding_box = [obj.matrix_world * Vector(corner) for corner in obj.bound_box]
    lowest_corner = None
    for corner in bounding_box:
        if lowest_corner:
            if corner.z < lowest_corner.z:
                lowest_corner = corner
                break
        else:
            lowest_corner = corner
    return [corner for corner in bounding_box if corner.z == lowest_corner.z]


def get_opposite_corners(obj=None, verts=None, inverse=False):
    """
    This function will return the most extreme corners in
    a bounding square

    Variables:
    :param obj: You can pass in an object to calculate the extreme corners of the bottom bounds
    :param verts: an iterable containing the verts to pick from
    :param inverse: This will return the opposite corners

    :return a tuple containing the right and left opposite corners as Vector objects
    """
    if not obj and not verts:
        return None
    elif obj:
        box = get_bottom_bounds(obj)
    else:
        box = [vertex.co for vertex in verts]
    left = None
    right = None
    for corner in box:
        if left and right:
            if not inverse:
                if corner.x < left.x or corner.y < left.y:
                    left = corner
                elif corner.x > right.x or corner.y > right.y:
                    right = corner
            else:
                if corner.x < left.x or corner.y > left.y:
                    left = corner
                elif corner.x > right.x or corner.y < right.y:
                    right = corner
        else:
            left = corner
            right = corner
    return left, right


# ------------------------------------------------------------ #
#                           TREES                              #
# -------------------------------------------------------------#
class AddTree(bpy.types.Operator):
    """
    Adds a tree generated with random elements at the cursors position.
    The properties of the operator determine what the tree will look like, see
    their descriptions for information on what they do.
    """
    bl_idname = "mesh.primitive_tree_add"
    bl_label = "Add Tree"
    bl_options = {'REGISTER', 'UNDO'}

    # Property variables
    seed = IntProperty(
        name="Seed",
        description="This is the seed that is used to generate the tree",
        default=0)
    use_seed = BoolProperty(
        name="Use seeding",
        description="This is whether a non-random seed should be used",
        default=False)

    filepath = StringProperty(
        name="Leaf Texture Directory",
        description="This is the directory of textures that can be randomly picked from",
        subtype="DIR_PATH")

    spawn_leaves = BoolProperty(
        name="Spawn Leaves",
        description="Should the tree generator spawn leaves with the tree. "
                    "WARNING, this can heavily increase the time taken to generate tree",
        default=True)
    update = BoolProperty(
        name="Render Tree",
        description="Should the tree generator update the model whilst the options are being changed. "
                    "WARNING, this can heavily increase the time taken to generate tree",
        default=False)

    leaf_density = IntProperty(
        name="Leaf Density",
        description="This is the number of leaves per branch with 3 iterations left",
        min=0, max=10,
        default=4)
    leaf_size = FloatProperty(
        name="Leaf Size",
        description="This is the size of a single spawned leaf",
        min=0, max=10,
        default=0.25)
    leaf_bend = FloatProperty(
        name="Leaf Bend",
        description="This is how much the corners of the leaf should bend in",
        min=0, max=1,
        default=0.25)
    leaf_iteration_start = IntProperty(
        name="Leaf Iteration Start",
        description="This is the iteration in the recursion that the leaves will start to be put on the tree",
        min=0, max=10,
        default=3)

    trunk_length = FloatProperty(
        name="Trunk Length",
        description="This is the height of the trunk",
        min=0, max=100,
        default=2.4)
    trunk_radius = FloatProperty(
        name="Trunk Radius",
        description="This is the radius of the trunk",
        min=0, max=100,
        default=0.15)
    trunk_position_deviation = FloatProperty(
        name="T-Pos Deviation",
        description="This is the percentage(0-1) that the non-trunk-top branches will be randomly moved by.",
        min=0, max=1,
        default=0.5)

    density = IntProperty(
        name="Density",
        description="This is the number of branches that will come from the trunk",
        min=1, max=10,
        default=5)

    angle = FloatProperty(
        name="Angle",
        description="This is the angle in radians of the branches rotation in Y axis",
        min=0, max=radians(90),
        subtype="ANGLE",
        default=radians(55))
    angle_delta = FloatProperty(
        name="Angle Delta",
        description="This is the change in angle that will take place each iteration",
        min=0, max=1,
        default=0.1)
    angle_deviation = FloatProperty(
        name="Angle Deviation",
        description="This is the percentage that the angle can deviate by",
        min=0, max=2,
        default=0.6)

    depth = FloatProperty(
        name="Depth",
        description="This is the length of a branch",
        min=0, max=100,
        default=1.5)
    depth_delta = FloatProperty(
        name="Depth Delta",
        description="This is the change in depth that will take place each iteration",
        min=0, max=1,
        default=0.3)
    depth_deviation = FloatProperty(
        name="Depth Deviation",
        description="This is the percentage that the branch's depth can deviate by",
        min=0, max=2,
        default=0.5)

    radius = FloatProperty(
        name="Radius",
        description="This is the thickness of the branch",
        min=0, max=100,
        default=0.09)
    radius_delta = FloatProperty(
        name="Radius Delta",
        description="This is the change in radius that will take place each iteration",
        min=0, max=1,
        default=0.25)
    radius_deviation = FloatProperty(
        name="Radius Deviation",
        description="This is the percentage that the radius can deviate by",
        min=0, max=2,
        default=0.3)

    iteration = IntProperty(
        name="Iterations",
        description="This is the number of iterations left",
        min=1, max=15,
        default=6)
    iteration_deviation = FloatProperty(
        name="Iteration Deviation",
        description="This is the percentage chance that the branch iterations will be cut short "
                    "before full completion ",
        min=0, max=1,
        default=0.5)

    # Non property variables for internal use
    leaf_materials = []
    leaf_templates = []

    @staticmethod
    def create_leaf_material(name, texture_image):
        """
        This function will load the textures into materials
        The material will also allow for transparency on the texture

        Variables:
        :param name: This is the name of the new material
        :param texture_image: This is the image that is going to be used for the material texture

        :return The new material
        """
        # Create the new leaf material
        material = bpy.data.materials.new(name=name)
        material.use_nodes = True

        # Remove the default shaders
        material.node_tree.nodes.remove(material.node_tree.nodes.get('Diffuse BSDF'))
        material_output = material.node_tree.nodes.get('Material Output')
        mix2 = material.node_tree.nodes.new('ShaderNodeMixShader')
        mix2.inputs[0].default_value = 0.15
        transparent2 = material.node_tree.nodes.new('ShaderNodeBsdfTransparent')
        mix = material.node_tree.nodes.new('ShaderNodeMixShader')
        diffuse = material.node_tree.nodes.new('ShaderNodeBsdfDiffuse')
        transparent = material.node_tree.nodes.new('ShaderNodeBsdfTransparent')
        texture_coord = material.node_tree.nodes.new('ShaderNodeTexCoord')
        texture = material.node_tree.nodes.new('ShaderNodeTexImage')
        texture.image = texture_image

        # Link the shaders
        material.node_tree.links.new(material_output.inputs[0], mix2.outputs[0])
        material.node_tree.links.new(mix2.inputs[1], mix.outputs[0])
        material.node_tree.links.new(mix2.inputs[2], transparent2.outputs[0])
        material.node_tree.links.new(mix.inputs[0], texture.outputs[1])
        material.node_tree.links.new(mix.inputs[1], transparent.outputs[0])
        material.node_tree.links.new(mix.inputs[2], diffuse.outputs[0])
        material.node_tree.links.new(diffuse.inputs[0], texture.outputs[0])
        material.node_tree.links.new(texture.inputs[0], texture_coord.outputs[2])
        return material

    @staticmethod
    def create_leaf(leaf_template, position=Vector((0, 0, 0)), material=None):
        """
        This function create the leaf at the selected position with the
        material that has been provided as optional argument. The leaf should be
        created using a leaf template generated with the create_leaf_mesh function

        :param leaf_template: This is the template object that this leaf is based off
        :param position: This is the position that the leaf is created at
        :param material: this is the material that is used for the leaf
        :return:
        """
        leaf = leaf_template.copy()
        leaf.data = leaf_template.data.copy()
        bpy.context.scene.objects.link(leaf)
        leaf.location = position
        leaf.name = "procgen_leaf_part"
        if material:
            leaf.active_material = material
        return leaf

    def load_leaf_materials(self):
        """This function will load all of the leaf textures and return an array"""
        if not self.filepath:
            return []
        absolute_path = os.path.abspath(self.filepath)
        files = os.listdir(absolute_path)
        leaf_textures = [load_texture(absolute_path + '/' + file) for file in files]
        leaf_materials = [AddTree.create_leaf_material("LeafMaterial", texture) for texture in leaf_textures]
        return leaf_materials

    def create_leaf_mesh(self):
        """This function will generate a leaf mesh template from a plane with a random element to it"""
        # Add new plane and generate UVs
        # Can add calc_uvs=True for blender 1.77,
        # however the uvs/plane need flipping in x axis and lab computers is 1.76
        bpy.ops.mesh.primitive_plane_add(radius=self.leaf_size, rotation=(radians(180), 0, 0))
        bpy.ops.object.mode_set(mode='EDIT')
        bpy.ops.mesh.subdivide()
        bpy.ops.uv.smart_project()  # This is for blender version 1.76
        bpy.ops.object.mode_set(mode='OBJECT')

        # Set leafs information
        plane = bpy.context.active_object
        plane.name = 'procgen_leaf_mesh_template'

        # Get the corners
        vertices = plane.data.vertices
        corners = get_opposite_corners(verts=vertices)
        corners2 = get_opposite_corners(verts=vertices, inverse=True)
        corners_all = corners + corners2
        for corner in corners_all:
            corner.z -= uniform(0, self.leaf_bend)
        return plane

    def create_branch(self, position=Vector((0, 0, 0)), angle=radians(40),
                      depth=5, radius=0.2, spawn_leaves=True):
        """
        Create a branch from a cylinder primitive, the optional arguments supplied
        will determine the shape of the branch

        Variables:
        :param position: This is the starting position of the branch (the bottom)
        :param angle: This is the angle in radians of the branches rotation in Y axis
        :param depth: This is the length of a branch
        :param radius: This is the thickness of the branch
        :param spawn_leaves: This is whether the tree should also spawn leaves on its generation

        :return: The branch object
        """
        # Create branch
        top = copy.copy(position)
        top.z += depth
        difference = top - position
        create_cylinder(vertices=8,
                        location=position,
                        depth=depth,
                        radius=radius)
        branch = bpy.context.object
        branch.name = "procgen_tree_part"

        # Rotations and positions    
        rotation_z_amount = random() * (2 * pi)
        rotation_y = Quaternion((0, 1, 0), angle)
        rotation_z = Quaternion((0, 0, 1), rotation_z_amount)
        rotation = rotation_z * rotation_y  # Make a rotation in z axis then rotate in y axis
        euler = rotation.to_euler()

        # Rotate the difference vector to get top point
        offset = copy.copy(difference)
        offset.rotate(rotation)
        branch.rotation_euler = euler
        branch.location += offset / 2

        # Create the leaves
        if spawn_leaves:
            # Rotations for positioning the leaf texture like the branch
            leaf_rotation = rotation
            leaf_rotation *= Quaternion((0, 1, 0), radians(90))
            leaf_rotation *= Quaternion((0, 0, 1), radians(90))
            leaf_rotation *= Quaternion((1, 0, 0), radians(90))
            leaf_rotation *= Quaternion((0, 0, 1), radians(-90))

            # Create the required amount of random leaves
            for index in range(self.leaf_density):
                material = self.leaf_materials[randint(0, len(self.leaf_materials) - 1)]

                # The position of the leaf on the branch
                random_offset = copy.copy(difference)
                random_offset *= random()
                random_offset.x += self.leaf_size + radius

                # The random rotation of the leaf around the branch
                random_rotation_amount = uniform(-pi, pi)
                random_offset.rotate(Quaternion((0, 0, 1), random_rotation_amount))
                random_rotation = Quaternion((0, 0, 1), -random_rotation_amount)
                random_leaf_rotation = copy.copy(leaf_rotation)
                random_leaf_rotation *= random_rotation

                # Rotate like the branch
                random_offset.rotate(rotation)
                location = copy.copy(position)
                leaf_position = location + random_offset
                leaf_template = self.leaf_templates[randint(0, len(self.leaf_templates) - 1)]
                leaf = self.create_leaf(leaf_template, position=leaf_position, material=material)
                leaf_scale = uniform(0.8, 1)
                leaf.scale = (leaf_scale, leaf_scale, leaf_scale)
                leaf.rotation_euler = random_leaf_rotation.to_euler()
        return branch

    def binary_tree_recursive(self, position=Vector((0, 0, 0)), angle=radians(40), depth=5, radius=0.2, iteration=3):
        """
        This function will start the recursive binary tree
        process until the number of iterations have been complete
        this function can control a large amount of the look of the
        tree such as the change in the angle of the tree.

        :param position: This is the starting position of the branch (the bottom)
        :param angle: This is the angle in radians of the branches rotation in Y axis
        :param depth: This is the length of a branch
        :param radius: This is the thickness of the branch
        :param iteration: This is the number of iterations left

        :return the left and right branch in a tuple
        """
        # Base case
        if iteration < 1:
            return None

        # Check whether to spawn leaves
        if iteration <= self.iteration - self.leaf_iteration_start \
                and self.spawn_leaves \
                and len(self.leaf_materials) > 0:
            spawn_leaves = True
        else:
            spawn_leaves = False

        # Deviations, they are different for both the left and right branch
        left_depth = depth * (1 - (random() * self.depth_deviation))
        left_angle = angle * (1 - (random() * self.angle_deviation))
        left_radius = radius * (1 - (random() * self.radius_deviation))
        right_depth = depth * (1 - (random() * self.depth_deviation))
        right_angle = angle * (1 - (random() * self.angle_deviation))
        right_radius = radius * (1 - (random() * self.radius_deviation))
        iteration -= ceil(iteration * (random() * self.iteration_deviation))

        # Update the original variables
        angle *= (1 - self.angle_delta)
        depth *= (1 - self.depth_delta)
        radius *= (1 - self.radius_delta)

        # Left and right branch
        left_branch = self.create_branch(position=position, angle=-left_angle, depth=left_depth, radius=left_radius,
                                         spawn_leaves=spawn_leaves)
        right_branch = self.create_branch(position=position, angle=right_angle, depth=right_depth, radius=right_radius,
                                          spawn_leaves=spawn_leaves)

        # Left branch recursion and scaling
        position_new = get_cylinder_top(left_branch, position)
        branches = self.binary_tree_recursive(position=position_new, angle=angle, depth=depth, radius=radius,
                                              iteration=iteration)
        if branches:
            left, right = branches
            scale_objects_to_match(left_branch, left)

        # Right branch recursion and scaling
        position_new = get_cylinder_top(right_branch, position)
        branches = self.binary_tree_recursive(position=position_new, angle=angle, depth=depth, radius=radius,
                                              iteration=iteration)
        if branches:
            left, right = branches
            scale_objects_to_match(right_branch, left)
        return left_branch, right_branch

    def create_tree(self, position=Vector((0, 0, 0))):
        """
        This function will start the creation of the tree, it will first make
        a trunk which continues up into recursive binary branches. Then depending
        on the density entered, it will create other branches according to
        the other variables entered

        Variables:
        :param position: This is the starting position of the branch (the bottom)

        :return the tree object or the tree and leaves
        """
        # Trunk creation
        trunk = self.create_branch(position=position, angle=0, depth=self.trunk_length, radius=self.trunk_radius,
                                   spawn_leaves=False)
        trunk_top = get_cylinder_top(trunk, position)

        # Top branch creation
        branches = self.binary_tree_recursive(position=trunk_top, angle=self.angle, depth=self.depth,
                                              radius=self.radius,
                                              iteration=self.iteration)
        left, right = branches
        scale_objects_to_match(trunk, left)

        # Update variables
        radius_new = self.radius * (1 - self.radius_delta)
        radius_new *= (1 - (random() * self.radius_deviation))

        # Density trunk branches
        for split in range(self.density - 1):
            offset = trunk_top - position
            trunk_top_deviated = position + (offset * (1 - (random() * self.trunk_position_deviation)))
            branch = self.create_branch(position=trunk_top_deviated, angle=self.angle, depth=self.depth,
                                        radius=self.radius,
                                        spawn_leaves=False)

            # Update variables
            position_new = get_cylinder_top(branch, trunk_top_deviated)
            branches = self.binary_tree_recursive(position=position_new, angle=self.angle, depth=self.depth,
                                                  radius=radius_new,
                                                  iteration=self.iteration)
            if branches:
                left, right = branches
                scale_objects_to_match(branch, left)

        # Select, join and name
        tree = join_objects_by_name("procgen_tree_part")
        tree.name = "Tree"
        if "procgen_leaf_part" in bpy.context.scene.objects:
            leaves = join_objects_by_name("procgen_leaf_part")
            leaves.name = "Leaves"
            return tree, leaves
        return tree

    def execute(self, context):
        """This is the entry point for the tree creation"""
        bpy.context.scene.update()  # Needs updating if user changes render engine
        if self.update:
            if self.spawn_leaves and not bpy.context.scene.render.engine == 'CYCLES':
                self.report({'ERROR'}, "Must be in cycles render engine for leaf generation!")
                return {'CANCELLED'}
            cursor_position = context.scene.cursor_location
            if not self.use_seed:
                self.seed = time()
            seed(self.seed)

            # Make sure can use nodes for leaf textures
            self.leaf_materials = self.load_leaf_materials()
            self.leaf_templates = [self.create_leaf_mesh() for _ in range(10)]

            # Spawn tree at cursor position
            self.create_tree(position=cursor_position)

            # Remove the templates from the scene
            bpy.ops.object.select_all(action='DESELECT')
            for template in self.leaf_templates:
                bpy.context.scene.objects[template.name].select = True
            bpy.ops.object.delete()
        return {'FINISHED'}


# ------------------------------------------------------------ #
#                           FOREST                             #
# ------------------------------------------------------------ #
class AddForest(bpy.types.Operator):
    """Creates several trees at random locations on the bottom bounds of the currently selected object"""
    bl_idname = "mesh.primitive_forest_add"
    bl_label = "Add Forest"
    bl_options = {'REGISTER', 'UNDO'}

    population = IntProperty(
        name="Population",
        description="This is the number of trees in the forest",
        min=0, max=500,
        default=0)

    seed = IntProperty(
        name="Seed",
        description="This is the seed that is used to generate the tree",
        default=0)
    use_seed = BoolProperty(
        name="Use seeding",
        description="This is whether a non-random seed should be used",
        default=False)

    filepath = StringProperty(
        name="Leaf Texture Directory",
        description="This is the directory of textures that can be randomly picked from",
        subtype="DIR_PATH")

    spawn_leaves = BoolProperty(
        name="Spawn Leaves",
        description="Should the tree generator spawn leaves with the tree. "
                    "WARNING, this can heavily increase the time taken to generate tree",
        default=True)
    update = BoolProperty(
        name="Render Tree",
        description="Should the tree generator update the model whilst the options are being changed. "
                    "WARNING, this can heavily increase the time taken to generate tree",
        default=False)

    leaf_density = IntProperty(
        name="Leaf Density",
        description="This is the number of leaves per branch with 3 iterations left",
        min=0, max=10,
        default=4)
    leaf_size = FloatProperty(
        name="Leaf Size",
        description="This is the size of a single spawned leaf",
        min=0, max=10,
        default=0.25)
    leaf_bend = FloatProperty(
        name="Leaf Bend",
        description="This is how much the corners of the leaf should bend in",
        min=0, max=1,
        default=0.25)
    leaf_iteration_start = IntProperty(
        name="Leaf Iteration Start",
        description="This is the iteration in the recursion that the leaves will start to be put on the tree",
        min=0, max=10,
        default=3)

    trunk_length = FloatProperty(
        name="Trunk Length",
        description="This is the height of the trunk",
        min=0, max=100,
        default=2.4)
    trunk_radius = FloatProperty(
        name="Trunk Radius",
        description="This is the radius of the trunk",
        min=0, max=100,
        default=0.15)
    trunk_position_deviation = FloatProperty(
        name="T-Pos Deviation",
        description="This is the percentage(0-1) that the non-trunk-top branches will be randomly moved by.",
        min=0, max=1,
        default=0.5)

    density = IntProperty(
        name="Density",
        description="This is the number of branches that will come from the trunk",
        min=1, max=10,
        default=5)

    angle = FloatProperty(
        name="Angle",
        description="This is the angle in radians of the branches rotation in Y axis",
        min=0, max=radians(90),
        subtype="ANGLE",
        default=radians(55))
    angle_delta = FloatProperty(
        name="Angle Delta",
        description="This is the change in angle that will take place each iteration",
        min=0, max=1,
        default=0.1)
    angle_deviation = FloatProperty(
        name="Angle Deviation",
        description="This is the percentage that the angle can deviate by",
        min=0, max=2,
        default=0.6)

    depth = FloatProperty(
        name="Depth",
        description="This is the length of a branch",
        min=0, max=100,
        default=1.5)
    depth_delta = FloatProperty(
        name="Depth Delta",
        description="This is the change in depth that will take place each iteration",
        min=0, max=1,
        default=0.3)
    depth_deviation = FloatProperty(
        name="Depth Deviation",
        description="This is the percentage that the branch's depth can deviate by",
        min=0, max=2,
        default=0.5)

    radius = FloatProperty(
        name="Radius",
        description="This is the thickness of the branch",
        min=0, max=100,
        default=0.09)
    radius_delta = FloatProperty(
        name="Radius Delta",
        description="This is the change in radius that will take place each iteration",
        min=0, max=1,
        default=0.25)
    radius_deviation = FloatProperty(
        name="Radius Deviation",
        description="This is the percentage that the radius can deviate by",
        min=0, max=2,
        default=0.3)

    iteration = IntProperty(
        name="Iterations",
        description="This is the number of iterations that the tree should walk through recursively",
        min=1, max=15,
        default=6)
    iteration_deviation = FloatProperty(
        name="Iteration Deviation",
        description="This is the percentage chance that the branch iterations will be cut short "
                    "before full completion ",
        min=0, max=1,
        default=0.5)

    def execute(self, context):
        """The entry point for the forrest creation"""
        cursor_position = context.scene.cursor_location
        obj = bpy.context.active_object
        if not obj:
            self.report({'ERROR'}, "No object is selected!")
            return {'CANCELLED'}

        # Randomly spawn trees in a given range
        left, right = get_opposite_corners(obj)
        for entity in range(self.population):
            random_x = uniform(left.x, right.x)
            random_y = uniform(left.y, right.y)
            bpy.context.scene.cursor_location = Vector((random_x, random_y, left.z))
            bpy.ops.mesh.primitive_tree_add(seed=self.seed, use_seed=self.use_seed, filepath=self.filepath,
                                            spawn_leaves=self.spawn_leaves,
                                            update=self.update,
                                            leaf_density=self.leaf_density, leaf_size=self.leaf_size,
                                            leaf_bend=self.leaf_bend, leaf_iteration_start=self.leaf_iteration_start,
                                            trunk_length=self.trunk_length,
                                            trunk_radius=self.trunk_radius,
                                            trunk_position_deviation=self.trunk_position_deviation,
                                            density=self.density,
                                            angle=self.angle, angle_delta=self.angle_delta,
                                            angle_deviation=self.angle_deviation, depth=self.depth,
                                            depth_delta=self.depth_delta,
                                            depth_deviation=self.depth_deviation, radius=self.radius,
                                            radius_delta=self.radius_delta,
                                            radius_deviation=self.radius_deviation, iteration=self.iteration,
                                            iteration_deviation=self.iteration_deviation)
        context.scene.cursor_location = cursor_position
        return {'FINISHED'}


# ------------------------------------------------------------ #
#                           LAYOUTS                            #
# ------------------------------------------------------------ #
def draw_extras_tree(self, context):
    """
    This will layout the buttons in the create toolbar menu in object mode
    ObjectMode > Create > Tree/Forest
    """
    col = self.layout.column(align=True)
    col.label(text="Tree Extras:")
    col.operator("mesh.primitive_tree_add", text="Tree", icon='PLUGIN')
    col.operator("mesh.primitive_forest_add", text="Forest", icon='PLUGIN')
    return {'FINISHED'}


def draw_extras_tree_sub(self, context):
    """
    This will layout the buttons in the Add menu in the bottom toolbar
    Add > Mesh > Tree/Forest
    """
    self.layout.separator()
    self.layout.operator("mesh.primitive_tree_add", text="Tree", icon='PLUGIN')
    self.layout.operator("mesh.primitive_forest_add", text="Forest", icon='PLUGIN')
    return {'FINISHED'}


def register():
    """This function is called by blender when registering the add-ons"""
    bpy.utils.register_class(AddTree)
    bpy.utils.register_class(AddForest)
    bpy.types.VIEW3D_PT_tools_add_object.append(draw_extras_tree)
    bpy.types.INFO_MT_mesh_add.append(draw_extras_tree_sub)
    return {'FINISHED'}


def unregister():
    """This function is called by blender when un-registering the add-ons"""
    bpy.utils.unregister_class(AddTree)
    bpy.utils.unregister_class(AddForest)
    bpy.types.VIEW3D_PT_tools_add_object.remove(draw_extras_tree)
    bpy.types.INFO_MT_mesh_add.remove(draw_extras_tree_sub)
    return {'FINISHED'}


if __name__ == "__main__":
    register()
